@@ -9,6 +9,7 @@
 #include <sys/stat.h>
 #include <sys/syscall.h>
 #include <sys/types.h>
+#include <sys/uio.h>
 #include <sys/user.h>
 #include <sys/wait.h>
 #include <unistd.h>
@@ -19,9 +20,24 @@
 #include "utils.h"
 
 
-#define PROCESS_FREE        0
-#define PROCESS_ALLOCATED   1
-#define PROCESS_ATTACHED    2
+#ifndef __X32_SYSCALL_BIT
+#define __X32_SYSCALL_BIT 0x40000000
+#endif
+
+#ifndef NT_PRSTATUS
+#define NT_PRSTATUS 1
+#endif
+
+
+int trace_verbosity = 0;
+#define verbosity trace_verbosity
+
+
+#define PROCESS_FREE        0   /* unallocated entry in table */
+#define PROCESS_ALLOCATED   1   /* fork() done but not yet attached */
+#define PROCESS_ATTACHED    2   /* running process */
+#define PROCESS_UNKNOWN     3   /* attached but no corresponding fork() call
+                                 * has finished yet */
 
 struct Process {
     unsigned int identifier;
@@ -41,7 +57,7 @@ struct ExecveInfo {
     char **envp;
 };
 
-struct Process **processes;
+struct Process **processes = NULL;
 size_t processes_size;
 
 struct Process *trace_find_process(pid_t pid)
@@ -64,7 +80,25 @@ struct Process *trace_get_empty_process(void)
             return processes[i];
     }
 
+    /* Count unknown processes */
+    {
+        size_t unknown = 0;
+        for(i = 0; i < processes_size; ++i)
+            if(processes[i]->status == PROCESS_UNKNOWN)
+                ++unknown;
+        {
+            int many_unknown = unknown * 2 >= processes_size;
+            if(verbosity >= 2 || (verbosity >= 1 && many_unknown) )
+                fprintf(stderr, "%sthere are %u/%u UNKNOWN processes\n",
+                        many_unknown?"Warning: ":"",
+                        (unsigned int)unknown, (unsigned int)processes_size);
+        }
+    }
+
     /* Allocate more! */
+    if(verbosity >= 3)
+        fprintf(stderr, "Process table full (%d), reallocating\n",
+                (int)processes_size);
     {
         struct Process *pool;
         size_t prev_size = processes_size;
@@ -75,14 +109,38 @@ struct Process *trace_get_empty_process(void)
         {
             processes[i] = pool++;
             processes[i]->status = PROCESS_FREE;
-            processes[i]->in_syscall = 0;
-            processes[i]->current_syscall = -1;
-            processes[i]->syscall_info = NULL;
         }
         return processes[prev_size];
     }
 }
 
+void trace_count_processes(unsigned int *p_nproc, unsigned int *p_unknown)
+{
+    unsigned int nproc = 0, unknown = 0;
+    size_t i;
+    for(i = 0; i < processes_size; ++i)
+    {
+        switch(processes[i]->status)
+        {
+        case PROCESS_FREE:
+            break;
+        case PROCESS_UNKNOWN:
+            /* Exists but no corresponding syscall has returned yet */
+            ++unknown;
+        case PROCESS_ALLOCATED:
+            /* Not yet attached but it will show up eventually */
+        case PROCESS_ATTACHED:
+            /* Running */
+            ++nproc;
+            break;
+        }
+    }
+    if(p_nproc != NULL)
+        *p_nproc = nproc;
+    if(p_unknown != NULL)
+        *p_unknown = unknown;
+}
+
 int trace_add_files_from_proc(unsigned int process, pid_t pid,
                               const char *binary)
 {
@@ -105,7 +163,7 @@ int trace_add_files_from_proc(unsigned int process, pid_t pid,
      * bfe44000-bfe65000 rw-p 00000000 00:00 0          [stack]
      */
 
-#ifdef DEBUG
+#ifdef DEBUG_PROC_PARSER
     fprintf(stderr, "Parsing %s\n", procfile);
 #endif
     fp = fopen(procfile, "r");
@@ -127,7 +185,7 @@ int trace_add_files_from_proc(unsigned int process, pid_t pid,
                &inode,
                pathname);
 
-#ifdef DEBUG
+#ifdef DEBUG_PROC_PARSER
         fprintf(stderr,
                 "proc line:\n"
                 "    addr_start: %lx\n"
@@ -150,7 +208,7 @@ int trace_add_files_from_proc(unsigned int process, pid_t pid,
             if(strncmp(pathname, binary, 4096) != 0
              && strncmp(previous_path, pathname, 4096) != 0)
             {
-#ifdef DEBUG
+#ifdef DEBUG_PROC_PARSER
                 fprintf(stderr, "    adding to database\n");
 #endif
                 if(db_add_file_open(process, pathname, FILE_READ) != 0)
@@ -211,6 +269,12 @@ char *trace_unhandled_syscall(int syscall, struct Process *process)
     case SYS_utimes:
         name = "utimes";
         break;
+    case SYS_mq_open:
+        name = "mq_open";
+        break;
+    case SYS_mq_unlink:
+        name = "mq_unlink";
+        break;
 
     /* Path as second argument */
     case SYS_symlink:
@@ -242,9 +306,33 @@ char *trace_unhandled_syscall(int syscall, struct Process *process)
     case SYS_fchownat:
         name = "fchownat"; type = 2;
         break;
+    case SYS_faccessat:
+        name = "faccessat"; type = 2;
+        break;
     case SYS_readlinkat:
         name = "readlinkat"; type = 2;
         break;
+#ifdef SYS_fstatat
+    case SYS_fstatat:
+#endif
+#ifdef SYS_fstatat64
+    case SYS_fstatat64:
+#endif
+#ifdef SYS_newfstatat
+    case SYS_newfstatat:
+#endif
+        name = "fstatat"; type = 2;
+        break;
+
+    /* Others */
+    case SYS_ptrace:
+        name = "ptrace"; type = 2;
+        break;
+#ifdef SYS_name_to_handle_at
+    case SYS_name_to_handle_at:
+        name = "name_to_handle_at"; type = 2;
+        break;
+#endif
     }
 
     if(name == NULL)
@@ -277,19 +365,7 @@ int trace_handle_syscall(struct Process *process)
 {
     pid_t pid = process->pid;
     const int syscall = process->current_syscall;
-#ifdef DEBUG
-    if(syscall == SYS_open || syscall == SYS_execve || syscall == SYS_fork
-     || syscall == SYS_vfork || syscall == SYS_clone)
-    {
-        const char *callname = (syscall == SYS_open)?"open":
-                               (syscall == SYS_execve)?"execve":
-                               (syscall == SYS_fork)?"fork":
-                               (syscall == SYS_vfork)?"vfork":
-                               "clone";
-        fprintf(stderr, "syscall=%u %s, in_syscall=%u\n",
-                syscall, callname, process->in_syscall);
-    }
-#endif
+
     /* ********************
      * open(), creat(), access()
      */
@@ -304,14 +380,15 @@ int trace_handle_syscall(struct Process *process)
             pathname = abspath(process->wd, oldpath);
             free(oldpath);
         }
-#ifdef DEBUG
-        fprintf(stderr, "%s(\"%s\") = %d (%s)\n",
-                (syscall == SYS_open)?"open":
-                    (syscall == SYS_creat)?"creat":"access",
-                pathname,
-                (int)process->retvalue,
-                (process->retvalue >= 0)?"success":"failure");
-#endif
+        if(verbosity >= 3)
+        {
+            fprintf(stderr, "%s(\"%s\") = %d (%s)\n",
+                    (syscall == SYS_open)?"open":
+                        (syscall == SYS_creat)?"creat":"access",
+                    pathname,
+                    (int)process->retvalue,
+                    (process->retvalue >= 0)?"success":"failure");
+        }
         if(process->retvalue >= 0)
         {
             if(syscall == SYS_access)
@@ -355,20 +432,21 @@ int trace_handle_syscall(struct Process *process)
             pathname = abspath(process->wd, oldpath);
             free(oldpath);
         }
-#ifdef DEBUG
-        fprintf(stderr, "%s(\"%s\") = %d (%s)\n",
-                (syscall == SYS_stat
+        if(verbosity >= 3)
+        {
+            fprintf(stderr, "%s(\"%s\") = %d (%s)\n",
+                    (syscall == SYS_stat
 #ifdef SYS_stat64
-               || syscall == SYS_stat64
+                   || syscall == SYS_stat64
 #endif
 #ifdef SYS_oldstat
-               || syscall == SYS_oldstat
-#endif
-                 )?"stat":"lstat",
-                pathname,
-                (int)process->retvalue,
-                (process->retvalue >= 0)?"success":"failure");
+                   || syscall == SYS_oldstat
 #endif
+                     )?"stat":"lstat",
+                    pathname,
+                    (int)process->retvalue,
+                    (process->retvalue >= 0)?"success":"failure");
+        }
         if(process->retvalue >= 0)
         {
             if(db_add_file_open(process->identifier,
@@ -390,12 +468,13 @@ int trace_handle_syscall(struct Process *process)
             pathname = abspath(process->wd, oldpath);
             free(oldpath);
         }
-#ifdef DEBUG
-        fprintf(stderr, "readlink(\"%s\") = %d (%s)\n",
-                pathname,
-                (int)process->retvalue,
-                (process->retvalue >= 0)?"success":"failure");
-#endif
+        if(verbosity >= 3)
+        {
+            fprintf(stderr, "readlink(\"%s\") = %d (%s)\n",
+                    pathname,
+                    (int)process->retvalue,
+                    (process->retvalue >= 0)?"success":"failure");
+        }
         if(process->retvalue >= 0)
         {
             if(db_add_file_open(process->identifier,
@@ -417,11 +496,12 @@ int trace_handle_syscall(struct Process *process)
             pathname = abspath(process->wd, oldpath);
             free(oldpath);
         }
-#ifdef DEBUG
-        fprintf(stderr, "chdir(\"%s\") = %d (%s)\n", pathname,
-                (int)process->retvalue,
-                (process->retvalue >= 0)?"success":"failure");
-#endif
+        if(verbosity >= 3)
+        {
+            fprintf(stderr, "chdir(\"%s\") = %d (%s)\n", pathname,
+                    (int)process->retvalue,
+                    (process->retvalue >= 0)?"success":"failure");
+        }
         if(process->retvalue >= 0)
         {
             free(process->wd);
@@ -443,9 +523,6 @@ int trace_handle_syscall(struct Process *process)
          *            char *const argv[],
          *            char *const envp[]); */
         struct ExecveInfo *execi = malloc(sizeof(struct ExecveInfo));
-#ifdef DEBUG
-        fprintf(stderr, "Entering execve, getting arguments...\n");
-#endif
         execi->binary = tracee_strdup(pid, (void*)process->params[0]);
         if(execi->binary[0] != '/')
         {
@@ -455,26 +532,27 @@ int trace_handle_syscall(struct Process *process)
         }
         execi->argv = tracee_strarraydup(pid, (void*)process->params[1]);
         execi->envp = tracee_strarraydup(pid, (void*)process->params[2]);
-#ifdef DEBUG
-        fprintf(stderr, "Got arguments:\n  binary=%s\n  argv:\n",
-                execi->binary);
+        if(verbosity >= 3)
         {
-            /* Note: this conversion is correct and shouldn't need a cast */
-            const char *const *v = (const char* const*)execi->argv;
-            while(*v)
+            fprintf(stderr, "execve called:\n  binary=%s\n  argv:\n",
+                    execi->binary);
             {
-                fprintf(stderr, "    %s\n", *v);
-                ++v;
+                /* Note: this conversion is correct and shouldn't need a
+                 * cast */
+                const char *const *v = (const char* const*)execi->argv;
+                while(*v)
+                {
+                    fprintf(stderr, "    %s\n", *v);
+                    ++v;
+                }
+            }
+            {
+                size_t nb = 0;
+                while(execi->envp[nb] != NULL)
+                    ++nb;
+                fprintf(stderr, "  envp: (%u entries)\n", (unsigned int)nb);
             }
         }
-        {
-            /* Note: this conversion is correct and shouldn't need a cast */
-            size_t nb = 0;
-            while(execi->envp[nb] != NULL)
-                ++nb;
-            fprintf(stderr, "  envp: (%u entries)\n", (unsigned int)nb);
-        }
-#endif
         process->syscall_info = execi;
     }
     else if(process->in_syscall && syscall == SYS_execve)
@@ -482,18 +560,19 @@ int trace_handle_syscall(struct Process *process)
         struct ExecveInfo *execi = process->syscall_info;
         if(process->retvalue >= 0)
         {
-            /* Note: exec->argv needs cast to suppress a bogus GCC warning
+            /* Note: execi->argv needs a cast to suppress a bogus warning
              * While conversion from char** to const char** is invalid,
              * conversion from char** to const char*const* is, in fact, safe.
-             * G++ accepts it, GCC issues a warning */
+             * G++ accepts it, GCC issues a warning. */
             if(db_add_exec(process->identifier, execi->binary,
                            (const char *const*)execi->argv,
-                           (const char *const*)execi->envp) != 0)
+                           (const char *const*)execi->envp,
+                           process->wd) != 0)
                 return -1;
-#ifdef DEBUG
-            fprintf(stderr, "Proc %d successfully exec'd %s\n",
-                    process->pid, execi->binary);
-#endif
+            if(verbosity >= 3)
+                fprintf(stderr, "Proc %d successfully exec'd %s\n",
+                        process->pid, execi->binary);
+            /* Process will get SIGTRAP with PTRACE_EVENT_EXEC */
             if(trace_add_files_from_proc(process->identifier, process->pid,
                                          execi->binary) != 0)
                 return -1;
@@ -515,21 +594,53 @@ int trace_handle_syscall(struct Process *process)
         {
             pid_t new_pid = process->retvalue;
             struct Process *new_process;
-#ifdef DEBUG
-            fprintf(stderr, "Process %d created by %d via %s\n",
-                    new_pid, process->pid,
-                    (syscall == SYS_fork)?"fork()":
-                    (syscall == SYS_vfork)?"vfork()":
-                    "clone()");
-#endif
-            new_process = trace_get_empty_process();
-            new_process->status = PROCESS_ALLOCATED;
-            new_process->pid = new_pid;
-            new_process->in_syscall = 0;
-            new_process->wd = strdup(process->wd);
-#ifdef DEBUG
-            fprintf(stderr, "WD = \"%s\"\n", new_process->wd);
-#endif
+            if(verbosity >= 2)
+                fprintf(stderr,
+                        "Process %d created by %d via %s\n"
+                        "    (working directory: %s)\n",
+                        new_pid, process->pid,
+                        (syscall == SYS_fork)?"fork()":
+                        (syscall == SYS_vfork)?"vfork()":
+                        "clone()",
+                        process->wd);
+
+            /* At this point, the process might have been seen by waitpid in
+             * trace() or not. */
+            new_process = trace_find_process(new_pid);
+            if(new_process != NULL)
+            {
+                /* Process has been seen before and options were set */
+                if(new_process->status != PROCESS_UNKNOWN)
+                {
+                    fprintf(stderr, "Critical: just created process that is "
+                            "already running (status=%d)\n",
+                            new_process->status);
+                    return -1;
+                }
+                new_process->status = PROCESS_ATTACHED;
+                new_process->wd = strdup(process->wd);
+                ptrace(PTRACE_SYSCALL, new_process->pid, NULL, NULL);
+                if(verbosity >= 2)
+                {
+                    unsigned int nproc, unknown;
+                    trace_count_processes(&nproc, &unknown);
+                    fprintf(stderr, "%d processes (inc. %d unattached)\n",
+                            nproc, unknown);
+                }
+            }
+            else
+            {
+                /* Process hasn't been seen before (syscall returned first) */
+                new_process = trace_get_empty_process();
+                new_process->status = PROCESS_ALLOCATED;
+                /* New process gets a SIGSTOP, but we resume on attach */
+                new_process->pid = new_pid;
+                new_process->in_syscall = 0;
+                new_process->wd = strdup(process->wd);
+            }
+
+            /* Parent will also get a SIGTRAP with PTRACE_EVENT_FORK */
+
             if(db_add_process(&new_process->identifier,
                               process->identifier,
                               process->wd) != 0)
@@ -539,7 +650,7 @@ int trace_handle_syscall(struct Process *process)
     /* ********************
      * Other syscalls that might be of interest but that we don't handle yet
      */
-    else if(process->in_syscall && process->retvalue >= 0)
+    else if(verbosity >= 1 && process->in_syscall && process->retvalue >= 0)
     {
         char *desc = trace_unhandled_syscall(syscall, process);
         if(desc != NULL)
@@ -565,9 +676,19 @@ int trace_handle_syscall(struct Process *process)
     return 0;
 }
 
+void trace_set_options(pid_t pid)
+{
+    ptrace(PTRACE_SETOPTIONS, pid, 0,
+           PTRACE_O_TRACESYSGOOD |  /* Adds 0x80 bit to SIGTRAP signals
+                                     * if paused because of syscall */
+           PTRACE_O_TRACECLONE |
+           PTRACE_O_TRACEFORK |
+           PTRACE_O_TRACEVFORK |
+           PTRACE_O_TRACEEXEC);
+}
+
 int trace(pid_t first_proc, int *first_exit_code)
 {
-    int nprocs = 0;
     for(;;)
     {
         int status;
@@ -576,73 +697,111 @@ int trace(pid_t first_proc, int *first_exit_code)
 
         /* Wait for a process */
         pid = waitpid(-1, &status, __WALL);
+        if(pid == -1)
+        {
+            perror("waitpid failed");
+            return -1;
+        }
         if(WIFEXITED(status))
         {
-            fprintf(stderr, "Process %d exited, %d processes remain\n",
-                    pid, nprocs-1);
-            if(pid == first_proc && first_exit_code != NULL)
+            unsigned int nprocs, unknown;
+            int exitcode;
+            if(WIFSIGNALED(status))
             {
-                if(WIFSIGNALED(status))
-                    /* exit codes are 8 bits */
-                    *first_exit_code = 0x0100 | WTERMSIG(status);
-                else
-                    *first_exit_code = WEXITSTATUS(status);
+                /* exit codes are 8 bits */
+                exitcode = 0x0100 | WTERMSIG(status);
+                if(verbosity >= 2)
+                    fprintf(stderr, "Process %d exited (signal %d), %d "
+                            "processes remain\n",
+                            pid, WTERMSIG(status), nprocs-1);
             }
+            else
+            {
+                exitcode = WEXITSTATUS(status);
+                if(verbosity >= 2)
+                    fprintf(stderr, "Process %d exited (code %d), %d "
+                            "processes remain\n",
+                            pid, WEXITSTATUS(status), nprocs-1);
+            }
+
+            if(pid == first_proc && first_exit_code != NULL)
+                *first_exit_code = exitcode;
             process = trace_find_process(pid);
             if(process != NULL)
             {
+                if(db_add_exit(process->identifier, exitcode) != 0)
+                    return -1;
                 free(process->wd);
                 process->status = PROCESS_FREE;
             }
-            --nprocs;
+            trace_count_processes(&nprocs, &unknown);
+            if(verbosity >= 2)
+                fprintf(stderr, "Process %d exited, %d processes remain\n",
+                        pid, (unsigned int)nprocs);
             if(nprocs <= 0)
                 break;
+            if(unknown >= nprocs)
+            {
+                fprintf(stderr, "Critical: only UNKNOWN processes remaining "
+                        "(%d)\n", (unsigned int)nprocs);
+                return -1;
+            }
             continue;
         }
 
         process = trace_find_process(pid);
         if(process == NULL)
         {
-            fprintf(stderr, "Warning: found unexpected process %d\n", pid);
+            if(verbosity >= 3)
+                fprintf(stderr, "Process %d appeared\n", pid);
             process = trace_get_empty_process();
-            process->status = PROCESS_ALLOCATED;
+            process->status = PROCESS_UNKNOWN;
             process->pid = pid;
             process->in_syscall = 0;
-            process->wd = get_p_wd(pid);
-            if(db_add_first_process(&process->identifier, process->wd) != 0)
-                return -1;
-        }
-        if(process->status != PROCESS_ATTACHED)
-        {
-#ifdef DEBUG
-            fprintf(stderr, "Allocating Process for %d\n", pid);
-#endif
-            process->status = PROCESS_ATTACHED;
-
-            fprintf(stderr, "Process %d attached\n", pid);
-            ++nprocs;
-            ptrace(PTRACE_SETOPTIONS, pid, 0,
-                   PTRACE_O_TRACESYSGOOD |  /* Adds 0x80 bit to SIGTRAP signals
-                                             * if paused because of syscall */
-                   PTRACE_O_TRACECLONE |
-                   PTRACE_O_TRACEFORK |
-                   PTRACE_O_TRACEVFORK);
-            ptrace(PTRACE_SYSCALL, pid, NULL, NULL);
+            process->wd = NULL;
+            trace_set_options(pid);
+            /* Don't resume, it will be set to ATTACHED and resumed when fork()
+             * returns */
             continue;
         }
+        else if(process->status == PROCESS_ALLOCATED)
+        {
+            process->status = PROCESS_ATTACHED;
 
-#ifdef DEBUG
-        if(!WIFSTOPPED(status))
-            fprintf(stderr, "Process %d is NOT stopped\n", pid);
-        else if( (WSTOPSIG(status) & 0x80) == 0)
-            fprintf(stderr, "Process %d is NOT stopped because of syscall "
-                    "(WSTOPSIG=%u)\n", pid, WSTOPSIG(status));
-#endif
+            if(verbosity >= 3)
+                fprintf(stderr, "Process %d attached\n", pid);
+            trace_set_options(pid);
+            ptrace(PTRACE_SYSCALL, pid, NULL, NULL);
+            if(verbosity >= 2)
+            {
+                unsigned int nproc, unknown;
+                trace_count_processes(&nproc, &unknown);
+                fprintf(stderr, "%d processes (inc. %d unattached)\n",
+                        nproc, unknown);
+            }
+            continue;
+        }
 
         if(WIFSTOPPED(status) && WSTOPSIG(status) & 0x80)
         {
             struct user_regs_struct regs;
-            ptrace(PTRACE_GETREGS, pid, NULL, &regs);
+            size_t len = 0;
+            /* Try to use GETREGSET first, since iov_len allows us to know if
+             * 32bit or 64bit mode was used */
+#ifdef PTRACE_GETREGSET
+            {
+                struct iovec iov;
+                iov.iov_base = &regs;
+                iov.iov_len = sizeof(regs);
+                if(ptrace(PTRACE_GETREGSET, pid, NT_PRSTATUS, &iov) == 0)
+                    len = iov.iov_len;
+            }
+            if(len == 0)
+#endif
+            /* GETREGSET undefined or call failed, fallback on GETREGS */
+            {
+                ptrace(PTRACE_GETREGS, pid, NULL, &regs);
+            }
 #if defined(I386)
             process->current_syscall = regs.orig_eax;
             if(process->in_syscall)
@@ -657,29 +816,74 @@ int trace(pid_t first_proc, int *first_exit_code)
                 process->params[5] = regs.ebp;
             }
 #elif defined(X86_64)
-            process->current_syscall = regs.orig_rax;
-            if(process->in_syscall)
-                process->retvalue = regs.rax;
+            /* TODO : handle i386 compat and x32 (currently just warns) */
+            /* If len is known (not 0) and not that of x86_64 registers,
+             * or if len is not known (0) and CS is 0x23 (not as reliable) */
+            if( (len != 0 && len != sizeof(regs))
+             || (len == 0 && regs.cs == 0x23))
+            {
+                if(verbosity >= 1)
+                    fprintf(stderr, "Warning: process %d made a syscall in "
+                            "i386 compat mode\n", pid);
+            }
             else
             {
-                process->params[0] = regs.rdi;
-                process->params[1] = regs.rsi;
-                process->params[2] = regs.rdx;
-                process->params[3] = regs.r10;
-                process->params[4] = regs.r8;
-                process->params[5] = regs.r9;
+                process->current_syscall = regs.orig_rax;
+                if(process->in_syscall)
+                    process->retvalue = regs.rax;
+                else
+                {
+                    if(process->current_syscall & __X32_SYSCALL_BIT)
+                    {
+                        if(verbosity >= 1)
+                            fprintf(stderr, "Warning: process %d made an x32 "
+                                    "syscall\n", pid);
+                    }
+                    process->params[0] = regs.rdi;
+                    process->params[1] = regs.rsi;
+                    process->params[2] = regs.rdx;
+                    process->params[3] = regs.r10;
+                    process->params[4] = regs.r8;
+                    process->params[5] = regs.r9;
+                }
             }
 #endif
             if(trace_handle_syscall(process) != 0)
                 return -1;
         }
-        /* Continue on SIGTRAP */
+        /* Handle signals */
         else if(WIFSTOPPED(status))
         {
-#ifdef DEBUG
-            fprintf(stderr, "Resuming %d on signal\n", pid);
-#endif
-            ptrace(PTRACE_SYSCALL, pid, NULL, NULL);
+            int signum = WSTOPSIG(status) & 0x7F;
+
+            /* Synthetic signal for ptrace event: resume */
+            if(signum == SIGTRAP && status & 0xFF0000)
+                ptrace(PTRACE_SYSCALL, pid, NULL, NULL);
+            else if(signum == SIGTRAP)
+            {
+                /* Probably doesn't happen? Then, remove */
+                fprintf(stderr, "NOT delivering SIGTRAP to %d\n",
+                        pid);
+                fprintf(stderr, "    waitstatus=0x%X\n", status);
+                ptrace(PTRACE_SYSCALL, pid, NULL, NULL);
+            }
+            /* Other signal, let the process handle it */
+            else
+            {
+                siginfo_t si;
+                if(verbosity >= 2)
+                    fprintf(stderr, "Process %d caught signal %d\n",
+                            pid, signum);
+                if(ptrace(PTRACE_GETSIGINFO, pid, 0, (long)&si) >= 0)
+                    ptrace(PTRACE_SYSCALL, pid, NULL, signum);
+                else
+                {
+                    /* Not sure what this is for */
+                    perror("    NOT delivering");
+                    if(signum != SIGSTOP)
+                        ptrace(PTRACE_SYSCALL, pid, NULL, NULL);
+                }
+            }
         }
     }
 
@@ -689,7 +893,6 @@ int trace(pid_t first_proc, int *first_exit_code)
 void cleanup(void)
 {
     size_t i;
-#ifdef DEBUG
     {
         size_t nb = 0;
         for(i = 0; i < processes_size; ++i)
@@ -700,7 +903,6 @@ void cleanup(void)
         fprintf(stderr, "Cleaning up, %u processes to kill...\n",
                 (unsigned int)nb);
     }
-#endif
     for(i = 0; i < processes_size; ++i)
     {
         if(processes[i]->status != PROCESS_FREE)
@@ -709,11 +911,12 @@ void cleanup(void)
             free(processes[i]->wd);
         }
     }
-    free(processes); /* FIXME : We still leak memory here */
 }
 
 void sigint_handler(int signo)
 {
+    if(verbosity >= 1)
+        fprintf(stderr, "Cleaning up on SIGINT\n");
     (void)signo;
     cleanup();
     exit(1);
@@ -721,23 +924,26 @@ void sigint_handler(int signo)
 
 void trace_init(void)
 {
-    size_t i;
-    struct Process *pool;
     signal(SIGCHLD, SIG_DFL);
-    processes_size = 16;
-    processes = malloc(processes_size * sizeof(*processes));
-    pool = malloc(processes_size * sizeof(*pool));
-    for(i = 0; i < processes_size; ++i)
-    {
-        processes[i] = pool++;
-        processes[i]->status = PROCESS_FREE;
-        processes[i]->in_syscall = 0;
-        processes[i]->current_syscall = -1;
-        processes[i]->syscall_info = NULL;
-        processes[i]->wd = NULL;
-    }
-
     signal(SIGINT, sigint_handler);
+
+    if(processes == NULL)
+    {
+        size_t i;
+        struct Process *pool;
+        processes_size = 16;
+        processes = malloc(processes_size * sizeof(*processes));
+        pool = malloc(processes_size * sizeof(*pool));
+        for(i = 0; i < processes_size; ++i)
+        {
+            processes[i] = pool++;
+            processes[i]->status = PROCESS_FREE;
+            processes[i]->in_syscall = 0;
+            processes[i]->current_syscall = -1;
+            processes[i]->syscall_info = NULL;
+            processes[i]->wd = NULL;
+        }
+    }
 }
 
 int fork_and_trace(const char *binary, int argc, char **argv,
@@ -749,10 +955,8 @@ int fork_and_trace(const char *binary, int argc, char **argv,
 
     child = fork();
 
-#ifdef DEBUG
-    if(child != 0)
+    if(child != 0 && verbosity >= 2)
         fprintf(stderr, "Child created, pid=%d\n", child);
-#endif
 
     if(child == 0)
     {
@@ -766,7 +970,7 @@ int fork_and_trace(const char *binary, int argc, char **argv,
         /* Execute the target */
         execvp(binary, args);
         perror("Couldn't execute the target command (execvp returned)");
-        return 1;
+        exit(1);
     }
 
     if(db_init(database_path) != 0)
@@ -779,14 +983,13 @@ int fork_and_trace(const char *binary, int argc, char **argv,
     {
         struct Process *process = trace_get_empty_process();
         process->status = PROCESS_ALLOCATED; /* Not yet attached... */
+        /* We sent a SIGSTOP, but we resume on attach */
         process->pid = child;
         process->in_syscall = 0;
         process->wd = get_wd();
 
-        fprintf(stderr, "Process %d created by initial fork()\n", child);
-#ifdef DEBUG
-        fprintf(stderr, "WD = \"%s\"\n", process->wd);
-#endif
+        if(verbosity >= 2)
+            fprintf(stderr, "Process %d created by initial fork()\n", child);
         if(db_add_first_process(&process->identifier, process->wd) != 0)
         {
             cleanup();
